@@ -1,14 +1,14 @@
 ;;; ada-mode.el --- major-mode for editing Ada sources  -*- lexical-binding:t -*-
 ;;
-;; Copyright (C) 1994, 1995, 1997 - 2022  Free Software Foundation, Inc.
+;; Copyright (C) 1994, 1995, 1997 - 2023  Free Software Foundation, Inc.
 ;;
 ;; Author: Stephen Leake <stephen_leake@stephe-leake.org>
 ;; Maintainer: Stephen Leake <stephen_leake@stephe-leake.org>
 ;; Keywords: languages
 ;;  ada
-;; Version: 8.0.1
-;; package-requires: ((uniquify-files "1.0.1") (wisi "4.1.1") (gnat-compiler "1.0.0") (emacs "25.3"))
-;; url: http://www.nongnu.org/ada-mode/
+;; Version: 8.0.2
+;; package-requires: ((uniquify-files "1.0.1") (wisi "4.2.0") (gnat-compiler "1.0.1") (emacs "25.3"))
+;; url: https://www.nongnu.org/ada-mode/
 ;;
 ;; This file is part of GNU Emacs.
 ;;
@@ -23,7 +23,7 @@
 ;; GNU General Public License for more details.
 ;;
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 ;;
 ;;; Usage:
 ;;
@@ -46,15 +46,15 @@
 ;; `ada-compiler' selects which compiler to use; it can be overridden
 ;; by the "ada_compiler" setting in project files.
 ;;
-;; We also support a cross reference tool (also called xref tool) that
-;; is different from the compiler. For example, you can use a local
-;; GNAT compiler to generate and access cross-reference information,
-;; while using a cross-compiler for compiling the final
+;; We also support a cross reference backend (also called xref
+;; backend) that is different from the compiler. For example, you can
+;; use a local GNAT compiler to generate and access cross-reference
+;; information, while using a cross-compiler for compiling the final
 ;; executable. The user variable `ada-xref-backend' selects the xref
-;; tool; it can be overridden by the "xref_backend" setting in project
-;; files.
+;; backend; it can be overridden by the "xref_backend" setting in
+;; project files.
 ;;
-;; The indentation engine and skeleton tools are from the wisi
+;; The indentation engine and skeleton backends are from the wisi
 ;; package.
 ;;
 ;;; History:
@@ -117,7 +117,7 @@
 (defun ada-mode-version ()
   "Return Ada mode version."
   (interactive)
-  (let ((version-string "8.0.1"))
+  (let ((version-string "8.0.2"))
     (if (called-interactively-p 'interactive)
 	(message version-string)
       version-string)))
@@ -156,7 +156,7 @@ nil, only the file name."
   :type 'boolean
   :safe #'booleanp)
 
-;;;; keymap and menus
+;;;; keymaps and menus
 
 (defvar ada-mode-map
   (let ((map (make-sparse-keymap)))
@@ -166,25 +166,16 @@ nil, only the file name."
     (define-key map [return] 	 'wisi-case-adjust-interactive)
     (define-key map "\C-c`" 	 'ada-show-secondary-error)
     (define-key map "\C-c;"      (lambda () (error "use M-; instead"))) ; comment-dwim
-    (define-key map "\C-c<" 	 'ada-goto-declaration-start)
-    (define-key map "\C-c>" 	 'ada-goto-declaration-end)
     (define-key map "\C-c\M-`" 	 'wisi-fix-compiler-error)
     (define-key map "\C-c\C-a" 	 'ada-align)
     (define-key map "\C-c\C-b" 	 'ada-make-subprogram-body)
     (define-key map "\C-c\C-c"   'ada-build-make)
-    (define-key map "\C-c\C-d" 	 'wisi-goto-spec/body)
-    (define-key map "\C-c\M-d" 	 'wisi-show-declaration-parents)
     (define-key map "\C-c\C-e" 	 'wisi-skel-expand)
     (define-key map "\C-c\C-f" 	 'wisi-show-parse-error)
-    (define-key map "\C-c\C-i" 	 'wisi-indent-statement)
-    (define-key map [3 backtab] 'wisi-indent-containing-statement);; C-c backtab, translated from C-c S-tab
     (define-key map "\C-\M-i"    'completion-at-point)
-    (define-key map "\C-c\C-l" 	 'wisi-show-local-references)
     (define-key map "\C-c\C-m"   'ada-build-set-make)
-    (define-key map "\C-c\C-n" 	 'forward-sexp)
     (define-key map "\C-c\M-n" 	 'wisi-skel-next-placeholder)
     (define-key map "\C-c\C-o" 	 'ada-find-other-file)
-    (define-key map "\C-c\C-p" 	 'backward-sexp)
     (define-key map "\C-c\M-p" 	 'wisi-skel-prev-placeholder)
     (define-key map "\C-c\C-q" 	 'wisi-refresh-prj-cache)
     (define-key map "\C-c\C-r" 	 'wisi-show-references)
@@ -193,8 +184,6 @@ nil, only the file name."
     (define-key map "\C-c\C-t" 	 'ada-find-file)
     (define-key map "\C-c\C-v"   'ada-build-check)
     (define-key map "\C-c\C-w" 	 'wisi-case-adjust-at-point)
-    (define-key map "\C-c\C-x"   'wisi-show-overriding)
-    (define-key map "\C-c\M-x"   'wisi-show-overridden)
     (define-key map "\C-c\C-y" 	 'wisi-case-create-exception)
     (define-key map "\C-c\C-\M-y" 'wisi-case-create-partial-exception)
     (define-key map [C-down-mouse-3] 'ada-popup-menu)
@@ -205,8 +194,8 @@ nil, only the file name."
     (define-key map "\C-cr4" 'ada-refactor-4)
     (wisi-case-activate-keys map)
 
-    map
-  )  "Local keymap used for Ada mode.")
+    map)
+  "Local keymap used for Ada mode.")
 
 (defvar ada-mode-menu (make-sparse-keymap "Ada"))
 (easy-menu-define ada-mode-menu ada-mode-map "Menu keymap for Ada mode"
@@ -219,61 +208,50 @@ nil, only the file name."
     ["Customize"                  (customize-group 'ada)    t]
     ("Project files"
      ["Find and select project ..."   ada-build-prompt-select-prj-file t]
-     ["Show project"                  wisi-prj-show                    t]
-     ["Show project file search path" wisi-prj-show-prj-path           t]
-     ["Show source file search path"  wisi-prj-show-src-path           t]
-     ["Clear current project"         wisi-prj-clear-current           t]
+     ["Show project"                  wisi-prj-show 		       t]
+     ["Show project file search path" wisi-prj-show-prj-path 	       t]
+     ["Show source file search path"  wisi-prj-show-src-path 	       t]
+     ["Clear current project"         wisi-prj-clear-current 	       t]
     )
     ("Build"
-     ["Next compilation error"     next-error                t]
-     ["Show secondary error"       ada-show-secondary-error  t]
-     ["Fix compilation error"      wisi-fix-compiler-error   t]
-     ["Show last parse error"      wisi-show-parse-error     t]
-     ["Check syntax"               ada-build-check       t]
-     ["Show main"                  ada-build-show-main   t]
-     ["Build"                      ada-build-make        t]
-     ["Set main and Build"         ada-build-set-make    t]
-     ["Run"                        ada-build-run         t]
+     ["Next compilation error/xref"  next-error 	      t]
+     ["Show secondary error"         ada-show-secondary-error t]
+     ["Fix compilation error"        wisi-fix-compiler-error  t]
+     ["Show last parse error"        wisi-show-parse-error    t]
+     ["Check syntax"                 ada-build-check 	      t]
+     ["Show main"                    ada-build-show-main      t]
+     ["Build"                        ada-build-make 	      t]
+     ["Set main and Build"           ada-build-set-make       t]
+     ["Run"                          ada-build-run 	      t]
      )
     ("Navigate"
-     ["Other file"                    ada-find-other-file          t]
-     ["Find file in project"          project-find-file            t]
-     ["Goto declaration/body"         wisi-goto-spec/body          t]
-     ["Goto next statement keyword"   forward-sexp 			     wisi-parser-shared]
-     ["Goto prev statement keyword"   backward-sexp 			     wisi-parser-shared]
-     ["Goto subprogram/package start" ada-goto-declaration-start 	     wisi-parser-shared]
-     ["Goto subprogram/package end"   ada-goto-declaration-end 		     wisi-parser-shared]
-     ["Goto declarative region start" ada-goto-declarative-region-start      wisi-parser-shared]
-     ["Goto containing statement start" wisi-goto-containing-statement-start wisi-parser-shared]
-     ["Show parent declarations"        wisi-show-declaration-parents 	     (eq ada-xref-backend 'gpr_query)]
-     ["Show all references (classwide)" xref-find-references                 t]
-     ["Show all direct references"      wisi-show-references                 t]
-     ["Show local references"         wisi-show-local-references    (eq ada-xref-backend 'gpr_query)]
-     ["Show overriding"               wisi-show-overriding          (eq ada-xref-backend 'gpr_query)]
-     ["Show overridden"               wisi-show-overridden          (eq ada-xref-backend 'gpr_query)]
-     ["Goto prev position current buffer" pop-to-mark-command  t]
-     ["Goto prev position other buffer"   pop-global-mark      t]
-     ["Next placeholder"              wisi-skel-next-placeholder    t]
-     ["Previous placeholder"          wisi-skel-prev-placeholder    t]
+     ["Other file"                    ada-find-other-file        t]
+     ["Find file in project"          project-find-file          t]
+     ["Show all references (classwide)" xref-find-references     t]
+     ["Show all direct references"      wisi-show-references     t]
+     ["Goto prev position current buffer" pop-to-mark-command 	 t]
+     ["Goto prev position other buffer"   pop-global-mark 	 t]
+     ["Next placeholder"              wisi-skel-next-placeholder t]
+     ["Previous placeholder"          wisi-skel-prev-placeholder t]
      )
     ("Edit"
      ["Complete name at point"      completion-at-point     t]
      ["Expand skeleton"             wisi-skel-expand        t]
      ["Indent line or selection"    indent-for-tab-command  t]
      ["Indent current statement"    wisi-indent-statement   wisi-parser-shared]
-     ["Indent containing statement" wisi-indent-containing-statement    wisi-parser-shared]
-     ["Indent file"            (indent-region (point-min) (point-max))  t]
-     ["Align"                       ada-align               t]
-     ["Comment/uncomment selection" comment-dwim            t]
-     ["Fill comment paragraph"         ada-fill-comment-paragraph           t]
+     ["Indent containing statement" wisi-indent-containing-statement        wisi-parser-shared]
+     ["Indent file"            (indent-region (point-min) (point-max)) 	    t]
+     ["Align"                       ada-align 				    t]
+     ["Comment/uncomment selection" comment-dwim 			    t]
+     ["Fill comment paragraph"         ada-fill-comment-paragraph 	    t]
      ["Fill comment paragraph justify" (ada-fill-comment-paragraph 'full)   t]
      ["Fill comment paragraph postfix" (ada-fill-comment-paragraph 'full t) t]
-     ["Make body for subprogram"    ada-make-subprogram-body     wisi-parser-shared]
+     ["Make body for subprogram"    ada-make-subprogram-body                wisi-parser-shared]
      )
     ("Refactor"
      ["Method (Object) => Object.Method"   ada-refactor-1 wisi-parser-shared]
-     ["Object.Method   => Method (Object)" ada-refactor-2 wisi-parser-sharedt]
-     ["Element (Object, Index) => Object (Index)" ada-refactor-3 wisi-parser-sharedt]
+     ["Object.Method   => Method (Object)" ada-refactor-2 wisi-parser-shared]
+     ["Element (Object, Index) => Object (Index)" ada-refactor-3 wisi-parser-shared]
      ["Object (Index) => Element (Object, Index)" ada-refactor-4 wisi-parser-shared]
      )
     ("Casing"
@@ -290,6 +268,50 @@ nil, only the file name."
      ["Restart parser"                wisi-kill-parser            wisi-parser-shared]
      )))
 
+(defvar ada--statement-minor-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map "\C-c<" 	 'ada-goto-declaration-start)
+    (define-key map "\C-c>" 	 'ada-goto-declaration-end)
+    (define-key map "\C-c\C-i" 	 'wisi-indent-statement)
+    (define-key map [3 backtab] 'wisi-indent-containing-statement);; C-c backtab, translated from C-c S-tab
+    (define-key map "\C-c\C-n" 	 'forward-sexp)
+    (define-key map "\C-c\C-p" 	 'backward-sexp)
+
+    map)
+  "Local keymap used for Ada statement motion minor mode.")
+
+(defvar ada--statement-minor-mode-menu (make-sparse-keymap "Ada statement motion"))
+(easy-menu-define ada--statement-minor-mode-menu ada--statement-minor-mode-map
+  "Menu keymap for Ada statement motion minor mode"
+  '("Ada statement motion"
+    ["Goto next statement keyword"   forward-sexp 			     wisi-parser-shared]
+    ["Goto prev statement keyword"   backward-sexp 			     wisi-parser-shared]
+    ["Goto subprogram/package start" ada-goto-declaration-start 	     wisi-parser-shared]
+    ["Goto subprogram/package end"   ada-goto-declaration-end 		     wisi-parser-shared]
+    ["Goto declarative region start" ada-goto-declarative-region-start      wisi-parser-shared]
+    ["Goto containing statement start" wisi-goto-containing-statement-start wisi-parser-shared]
+    ))
+
+(defvar ada--gpr-query-minor-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map "\C-c\C-d" 	 'wisi-goto-spec/body) ;; uses xref backend
+    (define-key map "\C-c\M-d" 	 'wisi-show-declaration-parents)
+    (define-key map "\C-c\C-l" 	 'wisi-show-local-references)
+    (define-key map "\C-c\C-x"   'wisi-show-overriding)
+    (define-key map "\C-c\M-x"   'wisi-show-overridden)
+    map)
+  "Local keymap used for gpr query minor mode.")
+
+(defvar ada--gpr-query-minor-mode-menu (make-sparse-keymap "gpr-query"))
+(easy-menu-define gpr-query-menu ada--gpr-query-minor-mode-map "Menu keymap for gpr-query minor mode"
+  '("gpr-query"
+    ["Goto declaration/body"    wisi-goto-spec/body           t]
+    ["Show parent declarations" wisi-show-declaration-parents t]
+    ["Show local references"    wisi-show-local-references    t]
+    ["Show overriding"          wisi-show-overriding          t]
+    ["Show overridden"          wisi-show-overridden          t]
+    ))
+
 (easy-menu-define ada-context-menu nil
   "Context menu keymap for Ada mode"
   '("Ada"
@@ -691,9 +713,13 @@ Also sets `ff-function-name' for `ff-pre-load-hook'."
 	 (error ""))))
 
     (eglot
-     (user-error "als version 22 does not support which-function")
-     ;; eglot/LSP maybe supports "which function" via DocumentSymbol, which
-     ;; ada_language_server version 22 does not support.
+     (if (eglot--server-capable :DocumentSymbol)
+	(progn
+	  ;; eglot/LSP maybe supports "which function" via DocumentSymbol, which
+	  ;; ada_language_server version 22 does not support.
+	  (display-warning 'ada "ada-mode which-function via eglot not yet implemented.")
+	  "")
+       "")
      )
 
     (none "")
@@ -714,7 +740,9 @@ Also sets `ff-function-name' for `ff-pre-load-hook'."
   (interactive)
   (and wisi-parser-shared
        (save-excursion
-	 (memq (wisi-cache-nonterm (wisi-goto-statement-start)) '(use_clause with_clause))
+	 (let ((start-cache (wisi-goto-statement-start)))
+	   (when start-cache
+	     (memq (wisi-cache-nonterm (wisi-goto-statement-start)) '(use_clause with_clause))))
 	 )))
 
 (defun ada-goto-subunit-name ()
@@ -854,7 +882,7 @@ the file name."
 (defun ada-show-secondary-error ()
   "Show the next secondary file reference in the compilation buffer.
 A secondary file reference is defined by text having text
-property `ada-secondary-error'.  These can be set by
+property `gnat-secondary-error'.  These can be set by
 compiler-specific compilation filters."
   (interactive)
 
@@ -871,14 +899,14 @@ compiler-specific compilation filters."
     ;; to ’ada-goto-source’ below; disable that temporarily.
     (let ((display-buffer-overriding-action nil))
       (pop-to-buffer next-error-last-buffer nil t)
-      (setq pos (next-single-property-change (point) 'ada-secondary-error))
+      (setq pos (next-single-property-change (point) 'gnat-secondary-error))
       (unless pos
 	;; probably at end of compilation-buffer, in new compile
 	(goto-char (point-min))
-	(setq pos (next-single-property-change (point) 'ada-secondary-error)))
+	(setq pos (next-single-property-change (point) 'gnat-secondary-error)))
 
       (when pos
-	(setq item (get-text-property pos 'ada-secondary-error))
+	(setq item (get-text-property pos 'gnat-secondary-error))
 	;; file-relative-name handles absolute Windows paths from
 	;; g++. Do this in compilation buffer to get correct
 	;; default-directory.
@@ -886,7 +914,7 @@ compiler-specific compilation filters."
 
 	;; Set point in compilation buffer past this secondary error, so
 	;; user can easily go to the next one.
-	(goto-char (next-single-property-change (1+ pos) 'ada-secondary-error)))
+	(goto-char (next-single-property-change (1+ pos) 'gnat-secondary-error)))
 
       (pop-to-buffer start-buffer nil t);; for windowing history
       )
@@ -952,7 +980,7 @@ compiler-specific compilation filters."
 subprogram, or task declaration point is currently in or just
 after.  For `beginning-of-defun-function'."
   (interactive)
-  (unless wisi-parser-shared ;; FIXME: define capabilities
+  (unless wisi-parser-shared
     (user-error "goto-declaration-* not supported by this parser; add use clause manually"))
   (push-mark)
   (ada-validate-enclosing-declaration t 'navigate)
@@ -1575,6 +1603,12 @@ Prompts with completion, defaults to filename at point."
 (defvar which-func-functions) ;; which-func.el
 (defvar which-func-non-auto-modes) ;; ""
 
+(defvar ada-other-backend-setup-function nil
+  "When non-nil, called with no args from ada-mode if any of
+`ada-face-backend',
+`ada-indent-backend', `ada-xref-backend' are set to other.
+See `ada-eglot-setup' in ada-eglot.el for a similar function.")
+
 ;;;###autoload
 (cl-defun ada-parse-require-process (&key wait)
   "Start the Ada parser in an external process, if not already started.
@@ -1592,6 +1626,14 @@ Unless WAIT, does not wait for parser to respond. Returns the parser object."
     (wisi-parse-require-process parser :nowait (not wait))
     parser))
 
+(define-minor-mode ada--statement-minor-mode
+  "Minor mode enabling statement motion using the wisi parser."
+  :lighter nil :interactive nil)
+
+(define-minor-mode ada--gpr-query-minor-mode
+  "Minor mode for enabling navigating sources gpr_query."
+  :lighter nil :interactive nil)
+
 ;;;###autoload
 (define-derived-mode ada-mode prog-mode "Ada"
   "The major mode for editing Ada code."
@@ -1612,13 +1654,6 @@ Unless WAIT, does not wait for parser to respond. Returns the parser object."
   ;; indent-region to use it for all indentation. See
   ;; ada-fill-comment-paragraph.
 
-  ;; AdaCore standard style (enforced by -gnaty) requires two spaces
-  ;; after '--' in comments; this makes it easier to distinguish
-  ;; special comments that have something else after '--'
-  (set (make-local-variable 'comment-padding) "  ")
-
-  (set (make-local-variable 'require-final-newline) t)
-
   (setq font-lock-defaults
 	'(ada-font-lock-keywords ;; keywords
 	  nil ;; keywords-only; font-lock does keywords, comment, string faces via regexp. parser does identifiers.
@@ -1670,41 +1705,23 @@ Unless WAIT, does not wait for parser to respond. Returns the parser object."
   (setq wisi-prj-parse-undefined-function #'ada-prj-parse-undefined)
   (setq wisi-xref-full-path ada-xref-full-path)
 
-  ;; We can't set buffer-local wisi-disable-parser before this,
-  ;; because this is the first function run with major-mode set to
-  ;; ada-mode. We don't encourage setting global or dir-local
-  ;; wisi-disable-parser because other modes may need a wisi parser
-  ;; (gpr-mode, wisitoken-grammar-mode).
-  (when (or (eq ada-xref-backend   'eglot)
-	    (eq ada-indent-backend 'eglot)
-	    (eq ada-face-backend   'eglot))
-    (require 'ada-eglot)
-    (when (fboundp 'ada-eglot-setup)
-      (ada-eglot-setup)))
-
-  ;; We want at least one warning for missing ada-process-parse-exec,
-  ;; to give users a reminder to build/install it. But also allow
-  ;; suppressing that warning for users using eglot; ada-eglot-setup
-  ;; sets wisi-disable-parser if the wisi parser is not needed.
+  ;; Startup cases:
   ;;
-  (unless wisi-disable-parser
-    (unless (executable-find ada-process-parse-exec)
-	(display-warning
-	 'ada
-	 (format "Ada parser exec '%s' not found; install it, or set ada-*-backend to eglot."
-		 ada-process-parse-exec))
-	(setq-local wisi-disable-parser t)))
-
-  ;; wisi-setup tolerates parser nil.
-  (wisi-setup
-   :indent-calculate '(ada-wisi-comment)
-   :post-indent-fail 'ada-wisi-post-indent-fail
-   :parser (unless wisi-disable-parser (ada-parse-require-process)))
-
-  (when wisi-parser-shared
-    (set (make-local-variable 'beginning-of-defun-function) #'ada-goto-declaration-start)
-    (set (make-local-variable 'end-of-defun-function) #'ada-goto-declaration-end))
-
+  ;; 1) Newbie or small project; no wisi project, no gpr
+  ;; file. ada-*-background set by default from presence of als,
+  ;; ada_mode_wisi_parse, gpr_query.  Start eglot in ada-mode,
+  ;; ada-mode-hook, or ada-mode-post-local-vars.
+  ;;
+  ;; 2) ada-*-background set in dir-local vars or by default, wisi
+  ;; project with or without gpr file declared via wisi-select-* in
+  ;; dir-local vars. Start eglot in wisi-select-prj.
+  ;;
+  ;; 3) ada-*-background, wisi project declared in Makefile or
+  ;; elsewhere before Ada files opened; start eglot in
+  ;; wisi-select-prj.
+  ;;
+  ;; In order to accommodate 1 and 2, we call ada-eglot-setup and wisi-setup in
+  ;; ada-mode-post-local-vars.
   (add-hook 'hack-local-variables-hook #'ada-mode-post-local-vars nil t)
   )
 
@@ -1718,6 +1735,8 @@ Unless WAIT, does not wait for parser to respond. Returns the parser object."
 
   (remove-hook 'hack-local-variables-hook #'ada-mode-post-local-vars)
 
+  ;;; Handle variables that can easily be file-specific.
+
   ;; fill-region-as-paragraph in ada-fill-comment-paragraph does not
   ;; call syntax-propertize, so set comment syntax on
   ;; ada-fill-comment-prefix. In post-local because user may want to
@@ -1732,6 +1751,58 @@ Unless WAIT, does not wait for parser to respond. Returns the parser object."
   (setq wisi-language-keywords ada-keywords)
   (setq wisi-case-keyword ada-case-keyword)
   (setq wisi-case-adjust-p-function #'ada-case-adjust-p)
+
+  ;;; See comment in ada-mode on startup cases for rationale on doing
+  ;;; ada-eglot-setup here.
+
+  (setq-local wisi-disable-parser nil)
+
+  (when (or (eq ada-diagnostics-backend 'eglot)
+	    (eq ada-face-backend        'eglot)
+	    (eq ada-indent-backend      'eglot)
+	    ;; ada-statement-backend cannot be 'eglot
+	    (eq ada-xref-backend        'eglot))
+    (require 'ada-eglot)
+    (when (fboundp 'ada-eglot-setup)
+      (ada-eglot-setup)))
+
+  (when (or (eq ada-diagnostics-backend 'other)
+	    (eq ada-face-backend        'other)
+	    (eq ada-indent-backend      'other)
+	    (eq ada-statement-backend   'other)
+	    (eq ada-xref-backend        'other))
+    (when ada-other-backend-setup-function
+      (funcall ada-other-backend-setup-function)))
+
+  ;; We want at least one warning for missing ada-process-parse-exec,
+  ;; to give users a reminder to build/install it. But avoid that
+  ;; warning for users using eglot; ada-eglot-setup sets
+  ;; wisi-disable-parser if the wisi parser is not needed.
+  ;;
+  (unless wisi-disable-parser
+    (unless (executable-find ada-process-parse-exec)
+	(display-warning
+	 'ada
+	 (format "Ada parser exec '%s' not found; install it, or set ada-*-backend to eglot."
+		 ada-process-parse-exec))
+	(setq-local wisi-disable-parser t)))
+
+  (when (eq ada-statement-backend 'wisi)
+    (ada--statement-minor-mode))
+
+  (when (eq ada-xref-backend 'gpr_query)
+    (ada--gpr-query-minor-mode))
+
+  ;; wisi-setup tolerates parser nil.
+  (wisi-setup
+   :indent-calculate '(ada-wisi-comment)
+   :post-indent-fail 'ada-wisi-post-indent-fail
+   :parser (unless wisi-disable-parser (ada-parse-require-process)))
+
+  (when wisi-parser-shared
+    (setq-local beginning-of-defun-function #'ada-goto-declaration-start)
+    (setq-local end-of-defun-function #'ada-goto-declaration-end))
+
   )
 
 (put 'ada-mode 'custom-mode-group 'ada)
